// File : callbacks.cc

#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include <stdio.h>
#include <unistd.h>
extern GtkWidget *Window;
extern GdkPixmap *pixmap_rouge;
extern GdkPixmap *pixmap_jaune;
extern GdkPixmap *pixmap_vide;

// Debut code de Puissance5.cc
#define HASH 1 
#define NBMAXPOS  30
#define NBMAXHASH 113
#define NBCFORCEE 20 

int alphaachange=0;
int betaachange=0;
int nalpha,nbeta;
int finish=0;  
int NIVEAU;
int coupmeutrier[8];
int coupmeutrierB[8];
char CNOIR;
char CBLANC;

int coup;
int a_joue;

class PositionType;

class Session;
class Position { 
 public:
  Position();
  void SetSession(Session* _ptrsession) { 
    ptrsession=_ptrsession;
  }
  void Raz();
  inline int  NbCoupsJouablesBlanc();
  inline unsigned char CoupLegalBlanc(int);
  inline int  NbCoupsJouablesNoir();
  inline unsigned char CoupLegalNoir(int);
  inline int HashCode();
  int Evaluation(const char); 
  
  int Eval(const PositionType &qq);

  void JoueBlanc(unsigned char c);
  void JoueNoir(unsigned char c);
  inline void Supprime(unsigned char c);
  inline void Set(Position* p);
  inline int Equal(Position& p);
  void Dump();
  void ChangeCouleur() { 
    for(int i=0;i<6;i++) { 
      unsigned char t=ligB[i]; ligB[i]=ligN[i]; ligN[i]=t; 
    }
  }
 public:
  /*-------------------------------
    Tableau de jeu : 6 lignes
                     7 colonnes
  -------------------------------*/
  Session* ptrsession;
  unsigned char ligB[6]; 
  unsigned char ligN[6];
  unsigned char nblibre[7];
};

//-- **********************************************************************
//-- **********************************************************************
class Arbre { 
public:
  Arbre(Position* PosInit);
  int ExploreAndReturnMin(const int NbNiveaux,const int,int,int);
  int ExploreAndReturnMax(const int NbNiveaux,const int,int,int);
  int ExploreForceeMin();
  int ExploreForceeMax();
  int NoteCalculee() { return(MeilleurNote); } 
  unsigned char  CoupCalcule()  { return(MeilleurCoup); }
  void CoupMeutrierNoir();
  void CoupMeutrierBlanc();
private:
  Position* Pos;
  int  niveau;
  int MeilleurNote;
  unsigned char MeilleurCoup;
};

class TableauDePositions { 
 public:
  TableauDePositions();

  void Reset() { 
    for(int h=0;h<NBMAXHASH;h++) { 
      for(int c=0;c<NBMAXPOS && PositionOk[h][c];c++) {
	PositionOk[h][c]=0;
	DernierePos[h]=0;
      }
    }
  }

  int Dedans(Position* p,int& h,int &c) { 
    h=p->HashCode();
    for(c=0;c<NBMAXPOS && PositionOk[h][c];c++) {
      if(p->Equal(TabPosition[h][c])) { 
	return(1);
      }
    }
    return(0);
  }
  
  int NotePosition(int h,int c) { 
    return(TabEvaluation[h][c]);
  }

  void Insere(Position *p,int n) { 
    int h=p->HashCode();
    if(DernierePos[h]<NBMAXPOS) { 
      TabPosition[h][DernierePos[h]].Set(p);
      TabEvaluation[h][DernierePos[h]]=n;
      PositionOk[h][DernierePos[h]]=1;
      DernierePos[h]++;
    }
  }

 private:
  Position TabPosition[NBMAXHASH][NBMAXPOS];
  int      TabEvaluation[NBMAXHASH][NBMAXPOS];
  int      DernierePos[NBMAXHASH];
  unsigned char PositionOk[NBMAXHASH][NBMAXPOS];
} TableauDeHashCode;




class Session { 
public:
  Session() { 
    GAIN=10000; GAINTEST=9000; 
    CASE4=10; CASE3=4; CASE5=4; CASE6=1; CASE2=1; CASE7=0; CASE1=0;
    LIG3_VIDE=500;
    LIG2_VIDE=20;
    LIG2_VIDEM1=0;
    niveau=5;
    nbc=0;
  }
  Session(int _GAIN,int _GAINTEST,
	  int _CASE1,int _CASE2,
	  int _CASE3,int _CASE4,
	  int _CASE5,int _CASE6,
	  int _CASE7,
	  int _LIG3_VIDE,
	  int _LIG2_VIDE,
	  int _LIG2_VIDEM1,
	  int _Joueur) { 
    GAIN=_GAIN; GAINTEST=_GAINTEST;
    CASE1=_CASE1;
    CASE2=_CASE2;
    CASE3=_CASE3;
    CASE4=_CASE4;
    CASE5=_CASE5;
    CASE6=_CASE6;
    CASE7=_CASE7;
    LIG3_VIDE=_LIG3_VIDE;
    LIG2_VIDE=_LIG2_VIDE;
    LIG2_VIDEM1=_LIG2_VIDEM1;
    Joueur=_Joueur;
    nbposana=0;
    nbposhash=0;
    niveau=5;
    nbc=0;
  }
  

  void Init() { 
    position.Raz();
    position.SetSession(this);
    NoteDernierCoup=0;
    nbc=0;
  }
  
  int Fini() { //-- 1 si le dernier coup a ete joue 
    if(position.NbCoupsJouablesBlanc()==0) { 
      return(1);
    } 
    if((NoteDernierCoup<=-GAIN) || (NoteDernierCoup>=GAIN)) { 
      return(1);
    }
    return(0);
  }
  
  int AGagne() { //-- 1 si le pgm a gagne , -1 si l'adversaire a gagne 0 si nulle
    if(NoteDernierCoup<=-GAIN) return(1);
    if(NoteDernierCoup>= GAIN) return(-1);
    return(0);
  }

  int Joue(int &err) {   //-- err=1 si fini 0 sinon
    err=0;
    Arbre A(&position);
#if HASH
    TableauDeHashCode.Reset();
#endif
    
    
    position.ptrsession->niveau=NIVEAU;
    if(nbc<4) position.ptrsession->niveau=3;
    if(nbc==4) position.ptrsession->niveau=NIVEAU;
    
    if(position.ptrsession->nbposana<20000) position.ptrsession->niveau+=2;
    if(position.ptrsession->nbposana>100000) {
      if(position.ptrsession->niveau>5) position.ptrsession->niveau-=2;
    }
    
    position.ptrsession->nbposana=position.ptrsession->nbposhash=0;
    alphaachange=betaachange=1;

    position.Dump();

    //-- Recherche du coup meurtrier  23 sept 98 
    A.CoupMeutrierNoir();
    A.CoupMeutrierBlanc();
    
    
    
    A.ExploreAndReturnMin(position.ptrsession->niveau,
			  position.ptrsession->niveau,
			  -position.ptrsession->GAIN,
			  position.ptrsession->GAIN);
    
    
    
    /*++
    printf("\nJoueur:%d nbc:%3d nalpha=%+5d  nbeta=%+5d les %c Jouent  %d  (%d)   niv:%d nbposana:%lu nbposhash:%lu",
	   Joueur,nbc,nalpha,nbeta,
	   (Joueur)? 'O' : '#',
	   (int)A.CoupCalcule(),
	   (int)A.NoteCalculee(),
	   position.ptrsession->niveau,position.ptrsession->nbposana,position.ptrsession->nbposhash);
	   */
    if(Joueur) { 
      CNOIR='#';
      CBLANC='O';
    }
    else {
      CNOIR='O';
      CBLANC='#';
    }
    
    position.JoueNoir(A.CoupCalcule());
    nbc++;
    position.Dump();
    int stopoupas=position.Evaluation(' ');

    NoteDernierCoup=stopoupas;
    
    
    if(stopoupas==position.ptrsession->GAIN|| stopoupas==-position.ptrsession->GAIN) { 
      ; //printf("\n ****** c'est fini *******\n");
      err=1;
    }  
    
    if((int)A.NoteCalculee()>=position.ptrsession->GAINTEST) { 
      ; //printf("\n *** le Joueur  Gagne ***\n");
    }
    if((int)A.NoteCalculee()<=-position.ptrsession->GAINTEST) { 
      ; //printf("\n *** Je Gagne ***\n");
    }
    return(A.CoupCalcule());
  }


  void OpposantJoue(int c) { 
    nbc++;
    position.JoueBlanc(c);
  }
  
  

public:
  long unsigned nbposana,nbposhash;
  int GAIN,GAINTEST,CASE1,CASE2,CASE3,CASE4,CASE5,CASE6,CASE7;
  int LIG3_VIDE,LIG2_VIDE,LIG2_VIDEM1;
  int Joueur;
  int niveau;
  int nbc;
  Position position;
  int NoteDernierCoup;
};

Session SBlanc;
Session SNoirs;

inline int Position::HashCode() { 
  int h=((ligB[0]+ligB[1]+ligB[2]+ligB[3]+ligB[4]+ligB[5])&127)
       *((ligN[0]+ligN[1]+ligN[2]+ligN[3]+ligN[4]+ligN[5])&127);
  h%=NBMAXHASH; 
  return((int)h);
}

void Position::Set(Position* p) { 
  ptrsession=p->ptrsession;
  
    ligB[0]=p->ligB[0];    
    ligB[1]=p->ligB[1];    
    ligB[2]=p->ligB[2];    
    ligB[3]=p->ligB[3];    
    ligB[4]=p->ligB[4];    
    ligB[5]=p->ligB[5];    

    ligN[0]=p->ligN[0];    
    ligN[1]=p->ligN[1];    
    ligN[2]=p->ligN[2];    
    ligN[3]=p->ligN[3];    
    ligN[4]=p->ligN[4];    
    ligN[5]=p->ligN[5];    

    nblibre[0]=p->nblibre[0];
    nblibre[1]=p->nblibre[1];
    nblibre[2]=p->nblibre[2];
    nblibre[3]=p->nblibre[3];
    nblibre[4]=p->nblibre[4];
    nblibre[5]=p->nblibre[5];
    nblibre[6]=p->nblibre[6];
}

inline int Position::Equal(Position& p) { 
  if(ligB[0] !=p.ligB[0]) return(0);
  if(ligN[0] !=p.ligN[0]) return(0);
  if(ligB[1] !=p.ligB[1]) return(0);
  if(ligN[1] !=p.ligN[1]) return(0);
  if(ligB[2] !=p.ligB[2]) return(0);
  if(ligN[2] !=p.ligN[2]) return(0);
  if(ligB[3] !=p.ligB[3]) return(0);
  if(ligN[3] !=p.ligN[3]) return(0);
  if(ligB[4] !=p.ligB[4]) return(0);
  if(ligN[4] !=p.ligN[4]) return(0);
  if(ligB[5] !=p.ligB[5]) return(0);
  if(ligN[5] !=p.ligN[5]) return(0);
  return(1);
}

const unsigned char Masque[]={ 1,2,4,8,16,32,64,128 };

void Position::Dump() {
  char buffer [20];
  GtkWidget *aWidget;

  int l,c;
 
#if 0
  printf("\n---------------\n 1 2 3 4 5 6 7");
  for(l=5;l>=0;l--) {
    printf("\n");
    for(c=0;c<7;c++) {
      if(ligB[l]&Masque[c]) {
        printf("|%c",CNOIR);
      }
      else if(ligN[l]&Masque[c]) {
        printf("|%c",CBLANC);
      }
      else printf("| ");
    }
    printf("|");
  }
  fflush (stdout);
#endif
 
  for(l=5;l>=0;l--) { 
    for(c=0;c<7;c++) { 
      sprintf (buffer, "P%d%d", l, c);
      aWidget = lookup_widget (GTK_WIDGET (Window), buffer);
 
      if(ligB[l]&Masque[c]) {
	gtk_pixmap_set (GTK_PIXMAP (aWidget), pixmap_rouge, 0);
      }
      else if(ligN[l]&Masque[c]) { 
	gtk_pixmap_set (GTK_PIXMAP (aWidget), pixmap_jaune, 0);
      }
      else {
	; //gtk_pixmap_set (GTK_PIXMAP (aWidget), pixmap_vide, 0);
      }
    }
  }
  while (gtk_events_pending()) gtk_main_iteration();

}

void Position::Raz() {  
  ligB[0]=ligB[1]=ligB[2]=ligB[3]=ligB[4]=ligB[5]=0;
  ligN[0]=ligN[1]=ligN[2]=ligN[3]=ligN[4]=ligN[5]=0;
  nblibre[0]=nblibre[1]=nblibre[2]=nblibre[3]=nblibre[4]=nblibre[5]=nblibre[6]=6;
}

Position::Position() { 
  Raz();
}

void Position::JoueBlanc(unsigned char c) { 
  c--;
  unsigned char m=Masque[c];

  nblibre[c]--;
  if((ligN[0]&m)==0 && (ligB[0]&m)==0) { ligB[0]|=m;return; } 
  if((ligN[1]&m)==0 && (ligB[1]&m)==0) { ligB[1]|=m;return; } 
  if((ligN[2]&m)==0 && (ligB[2]&m)==0) { ligB[2]|=m;return; } 
  if((ligN[3]&m)==0 && (ligB[3]&m)==0) { ligB[3]|=m;return; } 
  if((ligN[4]&m)==0 && (ligB[4]&m)==0) { ligB[4]|=m;return; } 
  if((ligN[5]&m)==0 && (ligB[5]&m)==0) { ligB[5]|=m;return; } 
}

void Position::JoueNoir(unsigned char c) {
  c--;
  unsigned char m=Masque[c];
  nblibre[c]--;
  if((ligN[0]&m)==0 && (ligB[0]&m)==0) { ligN[0]|=m; return; } 
  if((ligN[1]&m)==0 && (ligB[1]&m)==0) { ligN[1]|=m; return; } 
  if((ligN[2]&m)==0 && (ligB[2]&m)==0) { ligN[2]|=m; return; } 
  if((ligN[3]&m)==0 && (ligB[3]&m)==0) { ligN[3]|=m; return; } 
  if((ligN[4]&m)==0 && (ligB[4]&m)==0) { ligN[4]|=m; return; } 
  if((ligN[5]&m)==0 && (ligB[5]&m)==0) { ligN[5]|=m; return; } 
}



inline void Position::Supprime(unsigned char c) { 
  c--;
  unsigned char m=Masque[c];
  nblibre[c]++;
  if((ligB[5]&m) || (ligN[5]&m)) { ligB[5]&=~m;ligN[5]&=~m;return; } 
  if((ligB[4]&m) || (ligN[4]&m)) { ligB[4]&=~m;ligN[4]&=~m;return; } 
  if((ligB[3]&m) || (ligN[3]&m)) { ligB[3]&=~m;ligN[3]&=~m;return; } 
  if((ligB[2]&m) || (ligN[2]&m)) { ligB[2]&=~m;ligN[2]&=~m;return; } 
  if((ligB[1]&m) || (ligN[1]&m)) { ligB[1]&=~m;ligN[1]&=~m;return; } 
  if((ligB[0]&m) || (ligN[0]&m)) { ligB[0]&=~m;ligN[0]&=~m;return; } 
  return;
}

inline int Position::NbCoupsJouablesBlanc() {
  int n=0;
  if(nblibre[0]) n++;
  if(nblibre[1]) n++;
  if(nblibre[2]) n++;
  if(nblibre[3]) n++;
  if(nblibre[4]) n++;
  if(nblibre[5]) n++;
  if(nblibre[6]) n++;
  return(n);
}

inline int Position::NbCoupsJouablesNoir() {
  return(NbCoupsJouablesBlanc());
}

inline unsigned char Position::CoupLegalBlanc(int n) { 
  int c,m,i=0;
  c=coupmeutrier[1]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrier[2]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrier[3]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrier[4]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrier[5]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrier[6]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrier[7]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  return 0; //++
}


inline unsigned char  Position::CoupLegalNoir(int n) { 
  int c,m,i=0;
  c=coupmeutrierB[1]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrierB[2]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrierB[3]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrierB[4]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrierB[5]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrierB[6]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  c=coupmeutrierB[7]; m=Masque[c-1]; if(nblibre[c-1]) i++; if(i==n) return(c);
  return 0; //++
}

//-- **********************************************************************


class PositionType { 
public:
  unsigned char lc[6];
  PositionType() {
    Raz(); 
  } 
  void Raz() { 
    lc[0]=0;
    lc[1]=0;
    lc[2]=0;
    lc[3]=0;
    lc[4]=0;
    lc[5]=0;
  }
  void Set(int l,int c) { 
    lc[l]|=Masque[c];
  }
  
  void Save() { 
    
  }
  void Copy(PositionType &other) { 
    other.lc[0]=lc[0];
    other.lc[1]=lc[1];
    other.lc[2]=lc[2];
    other.lc[3]=lc[3];
    other.lc[4]=lc[4];
    other.lc[5]=lc[5];
  }
    
  int AndCompte(unsigned char lig[6]) const {
    unsigned v;
    int n=0;
    v=lc[0] & lig[0];  do { if(v&1) n++; v>>=1; } while(v);
    v=lc[1] & lig[1];  do { if(v&1) n++; v>>=1; } while(v);
    v=lc[2] & lig[2];  do { if(v&1) n++; v>>=1; } while(v);
    v=lc[3] & lig[3];  do { if(v&1) n++; v>>=1; } while(v);
    v=lc[4] & lig[4];  do { if(v&1) n++; v>>=1; } while(v);
    v=lc[5] & lig[5];  do { if(v&1) n++; v>>=1; } while(v);
    return(n);
  }
};

int NB_POSITIONS_TYPE=0;
PositionType Tab_Position_Type[70];

int Position::Evaluation(const char ajouelederniercoup) {
  ptrsession->nbposana++;
  int EvalTotal=0;
  for(int i=0;i<6;i++) { 
    if(ligB[i]&8)    EvalTotal+=ptrsession->CASE4;
    if(ligB[i]&16)   EvalTotal+=ptrsession->CASE5;
    if(ligB[i]&4)    EvalTotal+=ptrsession->CASE3;
    if(ligB[i]&2)    EvalTotal+=ptrsession->CASE2;
    if(ligB[i]&32)   EvalTotal+=ptrsession->CASE6;
  }
  for(int nbp=0; nbp<NB_POSITIONS_TYPE; nbp++) { 
    int evali=Eval(Tab_Position_Type[nbp]);
    if(evali>=ptrsession->GAINTEST) {       
      EvalTotal=ptrsession->GAIN;
      return(EvalTotal);
    }
    if(evali<=-ptrsession->GAINTEST) { 
      EvalTotal=-ptrsession->GAIN;
      return(EvalTotal);
    }
    EvalTotal+=evali;
  }  

  return(EvalTotal);
  
}

int Position::Eval(const PositionType &PT) { 
  
  int nbB=PT.AndCompte(ligB);
  if(nbB==4) { return(ptrsession->GAIN); }
  
  int nbN=PT.AndCompte(ligN); 
  if(nbB==3) { 
    if(nbN==0) { return(ptrsession->LIG3_VIDE);  }
    return(0);
  }
  
  if(nbN==4) { return(-ptrsession->GAIN); } 
  if(nbB==0) { 
    if(nbN==3) { return(-ptrsession->LIG3_VIDE); }
  }

  if(nbN==2 && nbB==0) { return(-ptrsession->LIG2_VIDE); } 
  if(nbB==2 && nbN==0) { return( ptrsession->LIG2_VIDE); }

  return(0);
}

class quatre { 
public:
  quatre() { } 
  void generateur() {
    int l,c; 
    for(l=0;l<6;l++) { 
      for(c=0;c<7;c++) { 
	lignehgbd(l,c);
	lignebghd(l,c);
      }
    }
    for(l=0;l<6;l++) { 
      for(c=0;c<7;c++) { 
	lignehor(l,c);
	lignevert(l,c);
      }
    }
  }


  void lignehor(int l,int c) { 
    if(c<4) {
      PositionType p;
      p.Set(l,c++); 
      p.Set(l,c++); 
      p.Set(l,c++); 
      p.Set(l,c++); 
      p.Copy(Tab_Position_Type[NB_POSITIONS_TYPE++]);
    }
  } 
  
  void lignevert(int l,int c) {
    if(l<3) { 
      PositionType p;
      p.Set(l++,c); 
      p.Set(l++,c); 
      p.Set(l++,c); 
      p.Set(l++,c); 
      p.Copy(Tab_Position_Type[NB_POSITIONS_TYPE++]);
    }   
  } 
  void lignebghd(int l,int c) { 
    if(c<4 && l<3) { 
      PositionType p;
      p.Set(l++,c++); 
      p.Set(l++,c++); 
      p.Set(l++,c++); 
      p.Set(l++,c++); 
      p.Copy(Tab_Position_Type[NB_POSITIONS_TYPE++]);
    }
  } 
  void lignehgbd(int l,int c) { 
    if(c<4 && l>2) { 
      PositionType p;
      p.Set(l--,c++); 
      p.Set(l--,c++); 
      p.Set(l--,c++); 
      p.Set(l--,c++); 
      p.Copy(Tab_Position_Type[NB_POSITIONS_TYPE++]);
    }
  } 
};

//-- **********************************************************************
//-- **********************************************************************

TableauDePositions::TableauDePositions()  { 
  for(int h=0;h<NBMAXHASH;h++) { 
    DernierePos[h]=0;
    for(int i=0;i<NBMAXPOS;i++) { 
      PositionOk[h][i]=0;
    }
  }
}

Arbre::Arbre(Position* PosInit) : Pos(PosInit) { MeilleurCoup=255; }

void Arbre::CoupMeutrierNoir() { //-- Pour les noirs (programme)
  coupmeutrier[1]=1;
  coupmeutrier[2]=2;
  coupmeutrier[3]=3;
  coupmeutrier[4]=4;
  coupmeutrier[5]=5;
  coupmeutrier[6]=6;
  coupmeutrier[7]=7;
  int coups[8];
  int ordre[8];
  ordre[1]=coupmeutrier[1];
  ordre[2]=coupmeutrier[2];
  ordre[3]=coupmeutrier[3];
  ordre[4]=coupmeutrier[4];
  ordre[5]=coupmeutrier[5];
  ordre[6]=coupmeutrier[6];
  ordre[7]=coupmeutrier[7];
  coups[1]=coups[2]=coups[3]=coups[4]=coups[5]=coups[6]=coups[7]=0;
  int n=Pos->NbCoupsJouablesNoir();
  int i;
  for(i=1;(i<=n);i++) { 
    unsigned char c=Pos->CoupLegalNoir(i);
    Pos->JoueNoir(c);
    int eval=Pos->Evaluation('N');
    if(coups[c]<=-Pos->ptrsession->GAINTEST) { 
      Pos->Supprime(c);
      continue;
    }
    if(eval <=-Pos->ptrsession->GAINTEST) { 
      coups[c]=-Pos->ptrsession->GAIN;
      Pos->Supprime(c);
      continue;
    }
    int n1=Pos->NbCoupsJouablesNoir();
    for(int i1=1;(i1<=n1);i1++) { 
      unsigned char c1=Pos->CoupLegalNoir(i1);
      Pos->JoueNoir(c1);
      int eval1=Pos->Evaluation('N');
      if(eval1<coups[c1]) { 
	coups[c1]=eval1;
      }
      
      if(eval1<=-Pos->ptrsession->GAINTEST) { 
	if(c1==c) eval1>>=1;
	coups[c1]=eval1;
	coups[c] =eval1-1;
	Pos->Supprime(c1);
	continue;
      }
      Pos->Supprime(c1);
    }
    Pos->Supprime(c);
  }

  coups[4]-=30; coups[5]-=20; coups[3]-=20; coups[2]-=10; coups[6]-=10;
  do { 
    n=0;
    for(i=1;i<7;i++) { 
      if(coups[i]>coups[i+1]) { 
	int t=ordre[i]; ordre[i]=ordre[i+1];ordre[i+1]=t;
	t=coups[i]; coups[i]=coups[i+1]; coups[i+1]=t;
	n=1;
      }
    }
  }
  while(n==1);
  
  //++printf("\n");
  for(i=1;i<=7;i++) { 
    coupmeutrier[i]=ordre[i];
  } 
  //++printf("\n");
  
}


void Arbre::CoupMeutrierBlanc() { //-- Pour les noirs (programme)
  coupmeutrierB[1]=1;
  coupmeutrierB[2]=2;
  coupmeutrierB[3]=3;
  coupmeutrierB[4]=4;
  coupmeutrierB[5]=5;
  coupmeutrierB[6]=6;
  coupmeutrierB[7]=7;
  int coups[8];
  int ordre[8];
  ordre[1]=coupmeutrierB[1];
  ordre[2]=coupmeutrierB[2];
  ordre[3]=coupmeutrierB[3];
  ordre[4]=coupmeutrierB[4];
  ordre[5]=coupmeutrierB[5];
  ordre[6]=coupmeutrierB[6];
  ordre[7]=coupmeutrierB[7];
  coups[1]=coups[2]=coups[3]=coups[4]=coups[5]=coups[6]=coups[7]=0;
  int n=Pos->NbCoupsJouablesBlanc();
  int i;
  for(i=1;(i<=n);i++) { 
    unsigned char c=Pos->CoupLegalBlanc(i);
    Pos->JoueBlanc(c);
    int eval=-Pos->Evaluation('B');
    if(coups[c]<=-Pos->ptrsession->GAINTEST) { 
      Pos->Supprime(c);
      continue;
    }
    if(eval <=-Pos->ptrsession->GAINTEST) { 
      coups[c]=-Pos->ptrsession->GAIN;
      Pos->Supprime(c);
      continue;
    }
    int n1=Pos->NbCoupsJouablesBlanc();
    for(int i1=1;(i1<=n1);i1++) { 
      unsigned char c1=Pos->CoupLegalBlanc(i1);
      Pos->JoueBlanc(c1);
      int eval1=-Pos->Evaluation('B');
      if(eval1<coups[c1]) { 
	coups[c1]=eval1;
      }
      
      if(eval1<=-Pos->ptrsession->GAINTEST) { 
	if(c1==c) eval1>>=1;
	coups[c1]=eval1;
	coups[c] =eval1-1;
	Pos->Supprime(c1);
	continue;
      }
      Pos->Supprime(c1);
    }
    Pos->Supprime(c);
  }




  coups[4]-=30; coups[5]-=20; coups[3]-=20; coups[2]-=10; coups[6]-=10;
  do { 
    n=0;
    for(i=1;i<7;i++) { 
      if(coups[i]>coups[i+1]) { 
	int t=ordre[i]; ordre[i]=ordre[i+1];ordre[i+1]=t;
	t=coups[i]; coups[i]=coups[i+1]; coups[i+1]=t;
	n=1;
      }
    }
  }
  while(n==1);
  
  //++printf("\n");
  for(i=1;i<=7;i++) { 
    coupmeutrierB[i]=ordre[i];
  } 
  //++printf("\n");
  
}

//-- Exploration d une variante a partir de la position courante

int Arbre::ExploreForceeMax() {     return(0);
  int p=Pos->Evaluation(' ');
  if(p>=Pos->ptrsession->GAINTEST) return(p);
  int n=Pos->NbCoupsJouablesBlanc();
  //-- On cherche un coup qui ne perd pas tout de suite 
  for(int i=1;(i<=n);i++) { 
    unsigned char c=Pos->CoupLegalBlanc(i);
    Pos->JoueBlanc(c);
    if(Pos->Evaluation(' ') >=Pos->ptrsession->GAINTEST) { 
      Pos->Supprime(c);
      return(Pos->ptrsession->GAIN);
    }
    int n1=Pos->NbCoupsJouablesNoir();
    int boncoup=-1;
    for(int i1=1;i1<=n1 && boncoup==-1;i1++) { 
      unsigned char c1=Pos->CoupLegalNoir(i1);
      Pos->JoueNoir(c1);
      if(Pos->Evaluation(' ') <=-Pos->ptrsession->GAINTEST) {
	Pos->Supprime(c1);
	continue;
      }
      Pos->Supprime(c1);      
      boncoup=c1;
    }
    if(boncoup==-1) {
      Pos->Supprime(c);
    }
    else { 
      int q=ExploreForceeMin();
      Pos->Supprime(c);
      return(q);
    }
  } 
  return(-Pos->ptrsession->GAIN);
}

int Arbre::ExploreForceeMin() { return(0);
  int p=Pos->Evaluation(' ');
  if(p<=-Pos->ptrsession->GAINTEST) return(p);
  int n=Pos->NbCoupsJouablesNoir();
  //-- On cherche un coup qui ne perd pas tout de suite 
  for(int i=1;(i<=n);i++) { 
    unsigned char c=Pos->CoupLegalNoir(i);
    Pos->JoueNoir(c);
    if(Pos->Evaluation(' ') <=-Pos->ptrsession->GAINTEST) { 
      Pos->Supprime(c);
      return(-Pos->ptrsession->GAIN);
    }
    int n1=Pos->NbCoupsJouablesBlanc();
    int boncoup=-1;
    for(int i1=1;i1<=n1 && boncoup==-1;i1++) { 
      unsigned char c1=Pos->CoupLegalBlanc(i1);
      Pos->JoueBlanc(c1);
      if(Pos->Evaluation(' ') >=Pos->ptrsession->GAINTEST) {
	Pos->Supprime(c1);
	continue;
      }
      Pos->Supprime(c1);
      boncoup=c1;
    }
    if(boncoup==-1) {
      Pos->Supprime(c);
    }
    else { 
      int q=ExploreForceeMax();
      Pos->Supprime(c);
      return(q);
    }
  }   
  return(Pos->ptrsession->GAIN);
}

  


//--************************************************************   PGM qui joue
//-- C'est les NOIRS qui JOUENT   ---> NOTE la + petite possible
int Arbre::ExploreAndReturnMin(const int NbNiveaux,
			       const int NivInit,
			       int alpha,
			       int beta) { 

  //-------- si on est en bout d arbre, on retourne l evaluation
  if(NbNiveaux<=0)  {    
    int p=Pos->Evaluation(' ');
    return(p);
  }
  /* ------- Deja Perdu ??? ------------------ */
  if(Pos->Evaluation(' ') >=Pos->ptrsession->GAINTEST) { 
    MeilleurNote=Pos->ptrsession->GAIN-(NivInit-NbNiveaux);
    MeilleurCoup=-1;  
    return(MeilleurNote); // Pos->ptrsession->GAIN
  }
  /* ------- Deja Gagne ??? ------------------ */
  if(Pos->Evaluation(' ') <=-Pos->ptrsession->GAINTEST) { 
    MeilleurNote=-Pos->ptrsession->GAIN+(NivInit-NbNiveaux);
    MeilleurCoup=-1;  
    return(MeilleurNote); // -Pos->ptrsession->GAIN
  }
  /*-------- Recherche de Coup Force ----------*/
  int n,i,hc1,hc2;
  unsigned char c;
  int tnote;

  //--
  n=Pos->NbCoupsJouablesNoir();

  if(n==0) { MeilleurCoup=-1; MeilleurNote=Pos->ptrsession->GAIN; return(Pos->ptrsession->GAIN); } 
  /* --------------------------------------------------
  if(n==1) { 
    MeilleurCoup=Pos->CoupLegalNoir(1);
    Pos->JoueNoir(MeilleurCoup);
    MeilleurNote=Pos->Evaluation('N');
    Pos->Supprime(MeilleurCoup);
    return(MeilleurNote);
  }
  -------------------------------------------------- */

  //-- Recherche de Pos->ptrsession->GAIN immediat  pour les noirs;
  MeilleurNote=-Pos->ptrsession->GAIN;
  for(i=1;(i<=n)&&(MeilleurNote<Pos->ptrsession->GAINTEST);i++) { 
    unsigned char c=Pos->CoupLegalNoir(i);
    Pos->JoueNoir(c);
    if(Pos->Evaluation('N') <=-Pos->ptrsession->GAINTEST) { 
      Pos->Supprime(c);
      MeilleurCoup=c;        
      MeilleurNote=-Pos->ptrsession->GAIN+(NivInit-NbNiveaux);
      return(MeilleurNote);
    }
    Pos->Supprime(c);
  }
  
  //--  Recherche de Pos->ptrsession->GAIN immediat  pour les blancs (adversaire)
  n=Pos->NbCoupsJouablesBlanc();
  for(i=1;(i<=n);i++) { 
    c=Pos->CoupLegalBlanc(i);
    Pos->JoueBlanc(c);
    if(Pos->Evaluation('B') >=Pos->ptrsession->GAINTEST) { 
      Pos->Supprime(c);
      //-- Est ce qu'une autre case gagne pour les blancs
      for(int au=i+1;au<=n;au++) { 
	int ac=Pos->CoupLegalBlanc(au);
	Pos->JoueBlanc(ac);
	if(Pos->Evaluation('B') >=Pos->ptrsession->GAINTEST) { 
	  //-- c est perdu de toutes les facons 	  
		 MeilleurCoup=c;
	  MeilleurNote=Pos->ptrsession->GAIN-(NivInit-NbNiveaux);
	  Pos->Supprime(ac);
	  return(MeilleurNote);
	}
	Pos->Supprime(ac);
      }
//-- Une seule case gagne pour les blancs, on force les noirs a la jouer
     MeilleurCoup=c;  
     Pos->JoueNoir(c);
     MeilleurNote=Pos->ptrsession->GAIN-(NivInit-NbNiveaux);
     Arbre arbre(Pos);
     tnote=arbre.ExploreAndReturnMax(NbNiveaux-1,NivInit,alpha,beta);
     if(tnote>=Pos->ptrsession->GAINTEST) { 
       tnote=Pos->ptrsession->GAIN-(NivInit-NbNiveaux); 
     }
     
     Pos->Supprime(c);
     if(tnote<MeilleurNote) { 
       
       MeilleurCoup=c; 
       MeilleurNote=tnote;
       
     }
     
     
     return(MeilleurNote);
    }
    Pos->Supprime(c);
  }
  
#if HASH
  if((NbNiveaux<=4)&&(NbNiveaux>1)) {
    if(TableauDeHashCode.Dedans(Pos,hc1,hc2)) { 
      tnote=TableauDeHashCode.NotePosition(hc1,hc2); 
      Pos->ptrsession->nbposhash++; 
      MeilleurCoup=Pos->CoupLegalBlanc(1);
      return(tnote);
    }
  }
#endif
n=Pos->NbCoupsJouablesNoir();
MeilleurNote=Pos->ptrsession->GAIN;
  for(i=1;(i<=n)&&(MeilleurNote>-Pos->ptrsession->GAINTEST);i++) { 
    if(MeilleurNote>=alpha) { 
      c=Pos->CoupLegalNoir(i);
      Pos->JoueNoir(c);
      Arbre arbre(Pos);
      
      if(MeilleurNote<beta) {
	beta=MeilleurNote;
	nbeta=beta;
	betaachange=1;
      }
      tnote=arbre.ExploreAndReturnMax(NbNiveaux-1,NivInit,alpha,beta);

      Pos->Supprime(c);
      if(tnote<MeilleurNote) { 
	MeilleurCoup=c; 
	MeilleurNote=tnote;
	
      } 
    }
  }
#if HASH
  if((NbNiveaux<=4)&&(NbNiveaux>1)) {
    if(MeilleurNote>-Pos->ptrsession->GAINTEST && MeilleurNote<Pos->ptrsession->GAINTEST) { 
      TableauDeHashCode.Insere(Pos,MeilleurNote); 
    }
  }
#endif

  if(MeilleurCoup>7) { 
    MeilleurCoup=Pos->CoupLegalBlanc(1);
  }

  return(MeilleurNote);
}


//-- C'est les BLANCS qui JOUENT   ---> NOTE la + grande possible
int Arbre::ExploreAndReturnMax(const int NbNiveaux,
			       const int NivInit,
			       int alpha,
			       int beta) { 
  //-------- si on est en bout d arbre, on retourne l evaluation
  if(NbNiveaux<=0)  {
    int p=Pos->Evaluation(' ');
    return(p);
  }  
  /* ------- Deja Perdu ??? ------------------ */
  if(Pos->Evaluation(' ') >=Pos->ptrsession->GAINTEST) { 
    MeilleurNote=Pos->ptrsession->GAIN-(NivInit-NbNiveaux);
    MeilleurCoup=-1;  
    return(MeilleurNote);
  }
  /* ------- Deja Gagne ??? ------------------ */
  if(Pos->Evaluation(' ') <=-Pos->ptrsession->GAINTEST) { 
    MeilleurNote=-Pos->ptrsession->GAIN+(NivInit-NbNiveaux);
    MeilleurCoup=-1;  
    return(MeilleurNote);
  }
  /*-------- Recherche de Coup Force ----------*/
  int n,i,hc1,hc2;
  unsigned char c;
  int tnote;
  //--
  n=Pos->NbCoupsJouablesNoir();

  if(n==0) { MeilleurCoup=-1; MeilleurNote=Pos->ptrsession->GAIN; return(Pos->ptrsession->GAIN); } 
  /* ---------------- Pas bon, il faut descendre tout en bas de l arbre
    if(n==1) { 
    MeilleurCoup=Pos->CoupLegalNoir(1);
    Pos->JoueNoir(MeilleurCoup);
    MeilleurNote=Pos->Evaluation('N');
    Pos->Supprime(MeilleurCoup);
    return(MeilleurNote);
  }
  ---------------------------------------- */

  /*-------- Recherche de Coup Force ----------*/
  n=Pos->NbCoupsJouablesNoir();

  if(n==0) { MeilleurCoup=-1; MeilleurNote=-Pos->ptrsession->GAIN; return(-Pos->ptrsession->GAIN); } 
  
  MeilleurNote=Pos->ptrsession->GAIN-(NivInit-NbNiveaux);
  for(i=1;(i<=n)&&(MeilleurNote<Pos->ptrsession->GAINTEST);i++) { 
    unsigned char c=Pos->CoupLegalNoir(i);
    Pos->JoueNoir(c);
    if(Pos->Evaluation('N') <=-Pos->ptrsession->GAINTEST) { 
      Pos->Supprime(c);
      MeilleurCoup=c; 
      Pos->JoueBlanc(c);

      MeilleurNote=-Pos->ptrsession->GAIN+(NivInit-NbNiveaux);
      Arbre arbre(Pos);
      tnote=arbre.ExploreAndReturnMin(NbNiveaux-1,NivInit,alpha,beta);
      Pos->Supprime(c);

      if(tnote>MeilleurNote) { 
	MeilleurCoup=c; 
	MeilleurNote=tnote;
      }
      return(MeilleurNote);
    }
    Pos->Supprime(c);
  }

#if HASH
  if((NbNiveaux<=4)&&(NbNiveaux>1)) { 
    if(TableauDeHashCode.Dedans(Pos,hc1,hc2)) { 
      tnote=TableauDeHashCode.NotePosition(hc1,hc2); 
      Pos->ptrsession->nbposhash++;
      if(MeilleurCoup>7) { 
	MeilleurCoup=Pos->CoupLegalBlanc(1);
      }
      return(tnote);
    }
  }
#endif

  n=Pos->NbCoupsJouablesBlanc();
  MeilleurNote=-Pos->ptrsession->GAIN+(NivInit-NbNiveaux);
  for(i=1;(i<=n)&&(MeilleurNote<Pos->ptrsession->GAINTEST);i++) { 
    if(MeilleurNote<=beta) {
      c=Pos->CoupLegalBlanc(i);
      Pos->JoueBlanc(c);
      
      Arbre arbre(Pos);
      
      if(MeilleurNote>alpha) { 
	alpha=MeilleurNote;
	alphaachange=1;
	nalpha=alpha;
      }
      
      tnote=arbre.ExploreAndReturnMin(NbNiveaux-1,NivInit,alpha,beta);
      Pos->Supprime(c);
      
      if(tnote>MeilleurNote) { 
	MeilleurCoup=c; 
	MeilleurNote=tnote;
      }
    }
  }
#if HASH
  if((NbNiveaux<=4)&&(NbNiveaux>1)) { 
    if(MeilleurNote>-Pos->ptrsession->GAINTEST && MeilleurNote<Pos->ptrsession->GAINTEST) { 
      TableauDeHashCode.Insere(Pos,MeilleurNote); 
    }
  }
#endif
  if(MeilleurCoup>7) { 
    MeilleurCoup=Pos->CoupLegalBlanc(1);
  }
  return(MeilleurNote);
}
// Fi,n code Puissance5.cc 
  


// Les callbacks
void InitPlay () {
  CNOIR='#';
  CBLANC='O';
  alphaachange=0;
  betaachange=0;

  NIVEAU = 5;
  NIVEAU=(NIVEAU&126)+1; 

  quatre q;   q.generateur();

  SBlanc.LIG3_VIDE=1000;
  SNoirs.LIG3_VIDE= 500;
  SBlanc.LIG2_VIDE= 100;
  SNoirs.LIG2_VIDE= 100;

  SBlanc.Init();
  SNoirs.Init();
  SBlanc.Joueur=0;
  SNoirs.Joueur=1;

  a_joue = 0;
  finish = 0;
}

void ColumnClic (GtkButton *button, gpointer  user_data) {
  if (a_joue == 0) {
    coup = GPOINTER_TO_INT ( user_data ); // On recupere le coup
    a_joue = 1; // On autorise la tache de fond a traiter le coup et a faire jouer le computer
  } 
}

/*
void ColumnClic (GtkButton *button, gpointer  user_data) {
  int coup = GPOINTER_TO_INT ( user_data );
  GtkWidget *aButton;
  int err;

  // On rend insensible les boutons 
  aButton = lookup_widget (GTK_WIDGET (button), "C1");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
  aButton = lookup_widget (GTK_WIDGET (button), "C2");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
  aButton = lookup_widget (GTK_WIDGET (button), "C3");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
  aButton = lookup_widget (GTK_WIDGET (button), "C4");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
  aButton = lookup_widget (GTK_WIDGET (button), "C5");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
  aButton = lookup_widget (GTK_WIDGET (button), "C6");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
  aButton = lookup_widget (GTK_WIDGET (button), "C7");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);

  if(SBlanc.Fini()) { 
    finish=1; 
    GtkWidget *TaGagne = create_TaGagne ();
    gtk_widget_show (TaGagne);  
    return; 
  } 

  SNoirs.OpposantJoue (coup);
  coup=SNoirs.Joue(err);
  if(err) { 
    finish=1; 
    GtkWidget *JaiGagne = create_JaiGagne ();
    gtk_widget_show (JaiGagne);    
    return; 
  } 

  if(SNoirs.Fini()) { 
    finish=1; 
    GtkWidget *TaGagne = create_TaGagne ();
    gtk_widget_show (TaGagne);
    return; 
  } 

  SBlanc.OpposantJoue(coup);

  // On rend sensible les boutons 
  aButton = lookup_widget (GTK_WIDGET (button), "C1");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (button), "C2");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (button), "C3");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (button), "C4");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (button), "C5");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (button), "C6");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (button), "C7");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
}
*/

void on_ItemRestart_activate (GtkMenuItem *menuitem, gpointer user_data) {
  SBlanc.Init();
  SNoirs.Init();
  SBlanc.Joueur=0;
  SNoirs.Joueur=1;

  finish = 0;

  // On rend sensible les boutons 
  GtkWidget *aButton;
  aButton = lookup_widget (GTK_WIDGET (Window), "C1");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (Window), "C2");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (Window), "C3");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (Window), "C4");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (Window), "C5");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (Window), "C6");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  aButton = lookup_widget (GTK_WIDGET (Window), "C7");
  gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
  
  // On reaffiche le board
  char buffer [20];
  GtkWidget *aWidget;
 
  for(int l=5;l>=0;l--) { 
    for(int c=0;c<7;c++) { 
      sprintf (buffer, "P%d%d", l, c);
      aWidget = lookup_widget (GTK_WIDGET (Window), buffer);
      gtk_pixmap_set (GTK_PIXMAP (aWidget), pixmap_vide, 0);
    }
  }

  a_joue = 0;
}

gint Idle_run (gpointer data) {
  if (a_joue == 1) { // Le joueur vient de cliquer un coup
    int err;
    // On rend insensible les boutons
    GtkWidget *aButton;
    aButton = lookup_widget (GTK_WIDGET (Window), "C1");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
    aButton = lookup_widget (GTK_WIDGET (Window), "C2");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
    aButton = lookup_widget (GTK_WIDGET (Window), "C3");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
    aButton = lookup_widget (GTK_WIDGET (Window), "C4");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
    aButton = lookup_widget (GTK_WIDGET (Window), "C5");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
    aButton = lookup_widget (GTK_WIDGET (Window), "C6");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
    aButton = lookup_widget (GTK_WIDGET (Window), "C7");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 0);
    while (gtk_events_pending()) gtk_main_iteration();

    if(SBlanc.Fini()) { 
      finish=1; 
      GtkWidget *TaGagne = create_TaGagne ();
      gtk_widget_show (TaGagne); 
      a_joue = 0;
      return 1; 
    } 
  
    // while (gtk_events_pending()) gtk_main_iteration();

    SNoirs.OpposantJoue (coup);
    coup=SNoirs.Joue(err);
    if(err) { 
      finish=1; 
      GtkWidget *JaiGagne = create_JaiGagne ();
      gtk_widget_show (JaiGagne); 
      a_joue = 0;
      return 1; 
    } 
 
    if(SNoirs.Fini()) { 
      finish=1; 
      GtkWidget *TaGagne = create_TaGagne ();
      gtk_widget_show (TaGagne);
      a_joue = 0;
      return 1; 
    } 
    
    SBlanc.OpposantJoue(coup);
    // On rend sensible les boutons 
    aButton = lookup_widget (GTK_WIDGET (Window), "C1");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
    aButton = lookup_widget (GTK_WIDGET (Window), "C2");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
    aButton = lookup_widget (GTK_WIDGET (Window), "C3");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
    aButton = lookup_widget (GTK_WIDGET (Window), "C4");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
    aButton = lookup_widget (GTK_WIDGET (Window), "C5");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
    aButton = lookup_widget (GTK_WIDGET (Window), "C6");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
    aButton = lookup_widget (GTK_WIDGET (Window), "C7");
    gtk_widget_set_sensitive (GTK_WIDGET (aButton), 1);
while (gtk_events_pending()) gtk_main_iteration();
    a_joue = 0;
  }

  return 1; /* C'est un callback. Il faut retourner 1 !!! */
}
